#include "tree_structurer.hpp"
#include <algorithm>
#include <filesystem>
#include <fstream>
#include <sstream>
#include <stdexcept>

namespace fs = std::filesystem;

bool TreeStructurer::should_ignore_dir(const std::string& dirname) {
    static const std::vector<std::string> ignore_list = {
        "build", "venv", "myenv", "dist", "node_modules", "CMakeFiles",
        ".git", ".idea", ".vscode", "__pycache__", "**pycache**"
    };

    if (!dirname.empty() && (dirname == "__main__.py" || dirname == "__init__.py")) {
        return false;
    }

    if (std::find(ignore_list.begin(), ignore_list.end(), dirname) != ignore_list.end()) {
        return true;
    }

    if (!dirname.empty()) {
        if (dirname[0] == '.' || dirname[0] == '_') {
            return true;
        }
        if (dirname.find("__") == 0 && dirname.find("__", 2) != std::string::npos) {
            return true;
        }
    }

    return false;
}

bool TreeStructurer::should_ignore_file(const std::string& filename) {
    static const std::vector<std::string> ignore_extensions = {
        ".pyc", ".pyo", ".pyd", ".so", ".dll", ".dylib",
        ".o", ".obj", ".a", ".lib"
    };

    if (!filename.empty() && (filename == "__main__.py" || filename == "__init__.py")) {
        return false;
    }

    if (!filename.empty() && (filename[0] == '.' || filename[0] == '_')) {
        return true;
    }

    fs::path path(filename);
    std::string ext = path.extension().string();
    if (std::find(ignore_extensions.begin(), ignore_extensions.end(), ext) != ignore_extensions.end()) {
        return true;
    }

    return false;
}

std::string TreeStructurer::get_relative_path(const fs::path& path, const fs::path& base) {
    fs::path rel = fs::relative(path, base);
    return rel.string();
}

std::vector<fs::path> TreeStructurer::get_filtered_paths(const fs::path& start) {
    std::vector<fs::path> paths;
    fs::directory_options options = fs::directory_options::skip_permission_denied;

    try {
        if (!fs::exists(start)) {
            throw std::runtime_error("Directory does not exist: " + start.string());
        }
        
        if (!fs::is_directory(start)) {
            throw std::runtime_error("Path is not a directory: " + start.string());
        }

        paths.push_back(start);

        bool is_empty = fs::directory_iterator(start) == fs::directory_iterator();
        if (is_empty) {
            throw std::runtime_error("Directory is empty: " + start.string());
        }

        for (const auto& entry : fs::recursive_directory_iterator(start, options)) {
            const auto& path = entry.path();

            bool should_skip = false;
            for (const auto& component : path) {
                if (should_ignore_dir(component.string())) {
                    should_skip = true;
                    break;
                }
            }
            if (should_skip) continue;

            if (entry.is_directory()) {
                if (!should_ignore_dir(path.filename().string())) {
                    paths.push_back(path);
                }
            } else {
                if (!should_ignore_file(path.filename().string())) {
                    paths.push_back(path);
                }
            }
        }
    } catch (const fs::filesystem_error& e) {
        throw std::runtime_error("Error accessing path: " + std::string(e.what()));
    }

    std::sort(paths.begin(), paths.end());
    return paths;
}

std::vector<std::string> TreeStructurer::get_directory_structure(const std::string& startpath) {
    std::vector<std::string> result;
    
    std::string normalized_path = startpath;
    if (startpath.substr(0, 2) == ".\\" || startpath.substr(0, 2) == "./") {
        normalized_path = startpath.substr(2);
    }
    
    fs::path start = normalized_path.empty() ? fs::current_path() : fs::path(normalized_path);

    try {
        auto paths = get_filtered_paths(start);
        if (paths.empty()) {
            throw std::runtime_error("No valid files or directories found in: " + start.string());
        }
        std::vector<bool> is_last_at_level(256, false);
        
        for (size_t i = 1; i < paths.size(); ++i) {
            const auto& path = paths[i];
            std::string rel_path = get_relative_path(path, start);
            
            int level = std::count(rel_path.begin(), rel_path.end(), fs::path::preferred_separator);

            bool is_last = true;
            for (size_t j = i + 1; j < paths.size(); ++j) {
                std::string next_rel_path = get_relative_path(paths[j], start);
                int next_level = std::count(next_rel_path.begin(), next_rel_path.end(), fs::path::preferred_separator);
                if (next_level == level) {
                    is_last = false;
                    break;
                }
                if (next_level < level) {
                    break;
                }
            }

            is_last_at_level[level] = is_last;

            std::string line;
            for (int j = 0; j < level; ++j) {
                if (j == level - 1) {
                    line += is_last ? "└── " : "├── ";
                } else {
                    line += is_last_at_level[j] ? "    " : "│   ";
                }
            }

            line += path.filename().string();
            if (fs::is_directory(path)) {
                line += "/";
            }
            result.push_back(line);
        }
    } catch (const fs::filesystem_error& e) {
        throw std::runtime_error("Failed to access directory: " + std::string(e.what()));
    }

    return result;
}

//
// UPDATED: Instead of processing the lines “on the fly,” we now build the tree and then create the project.
// This also allows the same parsing logic to be reused by create_structure_from_string().
//
void TreeStructurer::create_structure_from_file(const std::string& filepath, const std::string& target_path) {
    std::vector<std::string> lines = read_structure_file(filepath);
    validate_structure(lines);
    TreeNode root = build_tree_from_lines(lines);
    create_node(root, target_path);
}

void TreeStructurer::create_structure_from_string(const std::string& structure, const std::string& target_path) {
    std::istringstream iss(structure);
    std::vector<std::string> lines;
    std::string line;
    
    while (std::getline(iss, line)) {
        if (!line.empty()) {
            lines.push_back(line);
        }
    }
    
    validate_structure(lines);
    TreeNode root = build_tree_from_lines(lines);
    create_node(root, target_path);
}

//
// UPDATED: Now using a Unicode‑aware (token‑based) indent calculation.
//
void TreeStructurer::validate_structure(const std::vector<std::string>& lines) {
    if (lines.empty()) {
        throw std::runtime_error("Empty structure provided");
    }
    
    int prev_indent = -1;
    for (const auto& line : lines) {
        int current_indent = static_cast<int>(get_indent_level(line));
        if (current_indent > prev_indent + 1) {
            throw std::runtime_error("Invalid indentation level in structure");
        }
        prev_indent = current_indent;
    }
}

//
// UPDATED: Parse the indent level by looking for 4‑character tokens.
//
size_t TreeStructurer::get_indent_level(const std::string& line) {
    size_t level = 0;
    size_t pos = 0;
    // Look for “blank” or vertical bar segments (each 4 characters)
    while (pos + 4 <= line.size() && (line.compare(pos, 4, "    ") == 0 || line.compare(pos, 4, "│   ") == 0)) {
        level++;
        pos += 4;
    }
    // Then look for the branch indicator at this level
    if (pos + 4 <= line.size() && (line.compare(pos, 4, "└── ") == 0 || line.compare(pos, 4, "├── ") == 0)) {
        level++;
    }
    return level;
}

//
// UPDATED: Build a tree of nodes from the structure lines using the same token‐based parsing.
// The first (root) line is expected to be a directory (with a trailing '/').
TreeStructurer::TreeNode TreeStructurer::build_tree_from_lines(const std::vector<std::string>& lines) {
    if (lines.empty()) {
        throw std::runtime_error("Empty structure provided");
    }

    // Process the first line as the root.
    std::string first_line = lines[0];
    size_t pos = 0;
    // Skip any indent tokens (if any)
    while (pos + 4 <= first_line.size() && (first_line.compare(pos, 4, "    ") == 0 || first_line.compare(pos, 4, "│   ") == 0)) {
        pos += 4;
    }
    if (pos + 4 <= first_line.size() && (first_line.compare(pos, 4, "└── ") == 0 || first_line.compare(pos, 4, "├── ") == 0)) {
        pos += 4;
    }
    std::string root_name = first_line.substr(pos);
    if (root_name.empty() || root_name.back() != DIRECTORY_MARKER) {
        throw std::runtime_error("Root must be a directory (ending with '" + std::string(1, DIRECTORY_MARKER) + "')");
    }
    root_name.pop_back(); // Remove trailing '/'

    TreeNode root{root_name, false, {}};
    std::vector<TreeNode*> stack;
    stack.push_back(&root);

    // Process remaining lines.
    for (size_t i = 1; i < lines.size(); ++i) {
        const std::string& line = lines[i];
        if (line.empty()) continue;
        size_t pos = 0;
        size_t current_level = 0;
        // Process any “blank” or vertical indent segments (4 characters each)
        while (pos + 4 <= line.size() && (line.compare(pos, 4, "    ") == 0 || line.compare(pos, 4, "│   ") == 0)) {
            current_level++;
            pos += 4;
        }
        // Process the branch token if present
        if (pos + 4 <= line.size() && (line.compare(pos, 4, "└── ") == 0 || line.compare(pos, 4, "├── ") == 0)) {
            current_level++;
            pos += 4;
        }
        std::string name = line.substr(pos);
        if (name.empty()) continue;
        bool is_file = true;
        if (name.back() == DIRECTORY_MARKER) {
            is_file = false;
            name.pop_back();
        }
        name = sanitize_path(name);
        if (name.empty()) continue;

        // Adjust the stack to the proper level.
        while (stack.size() > current_level) {
            stack.pop_back();
        }
        if (stack.empty()) {
            throw std::runtime_error("Invalid indentation structure in the file.");
        }
        TreeNode new_node{name, is_file, {}};
        stack.back()->children.push_back(new_node);
        if (!is_file) {
            // For a directory, push the new node onto the stack.
            stack.push_back(&stack.back()->children.back());
        }
    }

    return root;
}

void TreeStructurer::create_node(const TreeNode& node, const std::filesystem::path& current_path) {
    std::filesystem::path new_path = current_path / node.name;
    
    if (node.is_file) {
        // Create file
        std::filesystem::path parent_path = new_path.parent_path();
        if (!std::filesystem::exists(parent_path)) {
            std::filesystem::create_directories(parent_path);
        }
        create_file(new_path);
    } else {
        // Create directory
        create_directory(new_path);
        // Recursively create children
        for (const auto& child : node.children) {
            create_node(child, new_path);
        }
    }
}

void TreeStructurer::create_directory(const std::filesystem::path& path) {
    if (!std::filesystem::exists(path)) {
        std::filesystem::create_directories(path);
    }
}

void TreeStructurer::create_file(const std::filesystem::path& path) {
    if (!std::filesystem::exists(path)) {
        std::ofstream file(path);
        if (!file.is_open()) {
            throw std::runtime_error("Failed to create file: " + path.string());
        }
        file.close();
    }
}

std::string TreeStructurer::sanitize_path(const std::string& path) {
    std::string result;
    for (char c : path) {
        // Allow alphanumeric, common punctuation, the directory separator, and our tree tokens.
        if (std::isalnum(static_cast<unsigned char>(c)) || 
            c == '.' || c == '_' || c == '-' || c == '/' || 
            c == TREE_PIPE || c == TREE_BRANCH || c == TREE_CORNER || c == TREE_DASH) {
            result += c;
        }
    }
    return result;
}

TreeStructurer::TreeNode TreeStructurer::parse_structure_line(const std::string& line, size_t indent_level) {
    size_t name_start = line.find_first_not_of(" │├└─");
    if (name_start == std::string::npos) {
        return TreeNode{"", true, {}};
    }

    std::string name = line.substr(name_start);
    name = sanitize_path(name);
    
    bool is_file = !name.empty() && name.back() != DIRECTORY_MARKER;
    if (!is_file) {
        name.pop_back();
    }
    
    return TreeNode{name, is_file, {}};
}

bool TreeStructurer::is_directory_marker(const std::string& line) {
    return !line.empty() && line.back() == DIRECTORY_MARKER;
}

std::string TreeStructurer::create_indent(size_t level) {
    return std::string(level * 4, ' ');
}

std::vector<std::string> TreeStructurer::read_structure_file(const std::string& filepath) {
    std::vector<std::string> lines;
    std::ifstream file(filepath);
    
    if (!file.is_open()) {
        throw std::runtime_error("Failed to open file: " + filepath);
    }
    
    std::string line;
    while (std::getline(file, line)) {
        if (!line.empty()) {
            lines.push_back(line);
        }
    }
    
    return lines;
}
